/* Clinch River 1D Routing Model
   Kinematic Wave Equation with temperature
   Developed in FORTRAN in 1987 by
   Dudley J. Benton */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdarg.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

void Abort(int line,char*format,...)
  {
  va_list arg_marker;
  va_start(arg_marker,format);
  fprintf(stderr,"\nerror %i: ",line);
  vfprintf(stderr,format,arg_marker);
  fprintf(stderr,"\n");
  _fcloseall();
  exit(1);
  }

double fmin(double x,double y)
  {
  if(x<y)
    return(x);
  return(y);
  }

double fmax(double x,double y)
  {
  if(x>y)
    return(x);
  return(y);
  }

/* river geometry */

#define ns 27 /* number of sections */

double s0[ns]={0.0008,0.0008,0.0008,0.0006,0.0003,0.00114,0.001,0.000833,
  0.00012,0.00012,0.00012,0.00013,0.00016,0.0003,0.0003,0.00033,0.00038,
  0.00042,0.0003,0.00018,0.00018,0.00018,0.00025,0.00035,0.0007,0.0007,0.0007};
double b[ns]={400.,400.,450.,380.,380.,430.,280.,250.,340.,320.,330.,330.,
  360.,340.,380.,445.,390.,340.,370.,350.,370.,400.,420.,520.,520.,450.,450.};
double rn[ns]={0.028,0.028,0.029,0.03,0.03,0.03,0.032,0.035,0.015,0.015,
  0.015,0.015,0.020,0.020,0.025,0.025,0.029,0.029,0.03,0.03,0.03,0.03,0.03,
  0.03,0.028,0.027,0.027};
double c[ns][2],h[ns][2],q[ns][2],rvbed[ns],tbed[ns][2],triv[ns][2];
double time;

double cbed=0.5;    /* specific heat of bed [BTU/lbm/F] */
double dbed=12.;    /* effective depth of bed [ft] */
double edifft=1E-5; /* eddy diffusivity [ft/sec] */
double htcsur=1.5;  /* surface heat transfer coefficient [BTU/hr/ft/F] */
double kbed=0.08;   /* thermal conductivity of bed [BTU/hr/ft/F] */
double kwater=0.4;  /* thermal conductivity of water [BTU/hr/ft/F] */
double rho=62.4;    /* water density [lbm/cu.ft.] */
double rhobed=106.; /* density of bed [lbm/cu.ft.] */
int new;

int Routing(double qin,double tin,double t0,int ndt,double*qout,double*tout)
  {
  int drybed,i,iter1,iter2,iter3;
  double c1,c11,c2,c3,cbar,change,cmax,cmx,dt,dt0,dx,ein,enew,eold,eout,
    hnew,htcbed,qb,qbar,qbxt,qinitl,qnew,rbedht,rvnew,rvold,s,surfht,thad,
    theta,tmax,tmin,tnew,v;
  static int initl=1;
  dt0=3600./ndt;
  dx=12.5*5280./(ns-1);
  if(initl!=0)
    {
    initl=0;
    qinitl=400.;
    v=qinitl/b[0]/pow(qinitl*rn[0]/(b[0]*pow(s0[0],0.5)*1.486),0.6);
    hnew=qinitl/v/b[0];
    for(i=0;i<ns;i++)
      {
      q[i][new]=qinitl;
      h[i][new]=hnew;
      c[i][new]=1.5*v;
      triv[i][new]=tin;
      tbed[i][new]=tin;
      rvbed[i]=dbed*rhobed*b[i]*dx;
      }
    }
  while(1)
    {
    new=1-new;
    for(i=0;i<ns;i++)
      {
      q[i][new]=q[i][1-new];
      h[i][new]=h[i][1-new];
      c[i][new]=c[i][1-new];
      triv[i][new]=triv[i][1-new];
      }
    iter1=0;
    cmx=0.;
    while(1)
      {
      iter1++;
      if(iter1>20)
        return(1);
      cmax=c[0][new];
      for(i=0;i<ns;i++)
        cmax=fmax(cmax,c[i][new]);
      q[0][new]=qin;
      triv[0][new]=tin;
      if(q[0][new]>0.)
        c[0][new]=1.5*q[0][new]/(b[0]*(q[0][new]*rn[0]/(b[0]*pow(s0[0],0.5)*1.486))*0.6);
      cmax=fmax(cmax,c[0][new]);
      dt=fmin(dt0,dx/cmax);
      for(i=1;i<ns;i++)
        {
        iter2=0;
        while(1)
          {
          iter2++;
          if(iter2>20)
            return(3);
          qbar=(q[i-1][1-new]+q[i][1-new])/2.;
          cbar=(c[i-1][new]+c[i][new]+c[i-1][1-new]+c[i][1-new])*0.25;
          qb=(q[i-1][new]+q[i][new])/2.;
          qbxt=(qbar+qb)/2.;
          theta=fmax(0.,fmin(0.9999,(1.-qbxt/(b[i]*s0[i]*cbar*dx))/2.));
          c11=dx/(cbar*dt);
          c3=exp(-cbar*dt/(dx*(1.-theta)));
          c1=1.-c11*(1.-c3);
          c2=c11*(1.-c3)-c3;
          qnew=q[i][new];
          q[i][new]=c1*q[i-1][new]+c2*q[i-1][1-new]+c3*q[i][1-new];
          if(q[i-1][new]>q[i][new]&&q[i][1-new]>q[i][new])
            q[i][new]=q[i][1-new];
          change=fabs(qnew-q[i][new]);
          if(iter2>6)
            q[i][new]=(q[i][new]+qnew)/2.;
          if(iter2>1&&change<50.)
            break;
          }
        }
      for(i=0;i<ns;i++)
        {
        v=0.1;
        if(q[i][new]>0.)
          {
          s=fmax(1.25e-4,(q[i][new]-q[i][1-new])/(dt*b[i]*pow(c[i][new],2)+s0[i]));
          h[i][new]=pow(q[i][new]*rn[i]/(1.486*b[i]*sqrt(s)),0.6);
          }
        v=q[i][new]/(b[i]*h[i][new]);
        c[i][new]=1.5*v;
        }
      change=fabs(cmax-cmx);
      if(iter1>1&&change<0.2)
        break;
      cmx=cmax;
      }
    for(i=1;i<ns;i++)
      {
      rvold=rho*h[i][1-new]*b[i]*dx;
      rvnew=rho*(q[i-1][1-new]+q[i-1][new]-q[i][1-new]-q[i][new])*dt/2.+rvold;
      drybed=0;
      if(rvnew<1000.)
        drybed=1;
      ein=rho*dt*(q[i-1][1-new]*triv[i-1][1-new]+q[i-1][new]*triv[i-1][new])/2.;
      eout=rho*dt*(q[i][1-new]*triv[i][1-new]+q[i][new]*triv[i][new])/2.;
      eold=triv[i][1-new]*rvold;
      tmin=fmin(triv[i-1][1-new],fmin(triv[i][1-new],fmin(triv[i-1][new],fmin(t0,tbed[i][1-new]))));
      tmax=fmax(triv[i-1][1-new],fmax(triv[i][1-new],fmax(triv[i-1][new],fmax(t0,tbed[i][1-new]))));
      iter3=0;
      while(1)
        {
        iter3++;
        if(iter3>20)
          return(3);
        change=0.;
        if(!drybed)
          {
          thad=edifft*(b[i-1]*h[i-1][1-new]*(triv[i-1][1-new]-triv[i][1-new])+b[i]*h[i][1-new]*(triv[min(ns,i+1)-1][1-new]-triv[i][1-new])+b[i-1]*h[i-1][new]*(triv[i-1][new]-triv[i][new])+b[i]*h[i][new]*(triv[min(ns,i+1)-1][1-new]-triv[i][new]))/2./dx;
          surfht=dx*b[i]*htcsur*(2.*t0-triv[i][new]-triv[i][1-new])/2./3600.;
          }
        if(drybed)
          {
          htcbed=htcsur;
          rbedht=dx*b[i]*htcbed*(tbed[i][new]+tbed[i][1-new]-2.*t0)/2./3600.;
          }
        else
          {
          htcbed=2./(h[i][new]/kwater+(dbed/kbed));
          rbedht=dx*b[i]*htcbed*(tbed[i][new]+tbed[i][1-new]-triv[i][new]-triv[i][1-new])/2./3600.;
          }
        enew=eold+ein-eout+(surfht+thad+rbedht)*dt;
        if(!drybed)
          tnew=enew/rvnew;
        tnew=fmax(tmin,fmin(tmax,tnew));
        change=fabs(triv[i][new]-tnew);
        triv[i][new]=(triv[i][new]+tnew)/2.;
        tnew=tbed[i][1-new]-rbedht*dt/rvbed[i]/cbed;
        tnew=fmax(tmin,fmin(tmax,tnew));
        change=fmax(change,fabs(tbed[i][new]-tnew));
        tbed[i][new]=(tbed[i][new]+tnew)/2.;
        if(drybed)
          triv[i][new]=tbed[i][new];
        if(iter3>1&&change<1.)
          break;
        }
      }
    time+=dt;
    if(time>=dt0)
      break;
    }
  time-=dt0;
  *qout=q[ns-1][1-new]+(q[ns-1][new]-q[ns-1][1-new])*(dt0-time)/dt0;
  drybed=0;
  if(*qout<10.)
    drybed=1;
  if(drybed)
    *tout=tbed[ns-1][1-new]+(tbed[ns-1][new]-tbed[ns-1][1-new])*(dt0-time)/dt0;
  else
    *tout=triv[ns-1][1-new]+(triv[ns-1][new]-triv[ns-1][1-new])*(dt0-time)/dt0;
  return(0);
  }

#define ndt 5 /* number of time steps per hour */

int main(int argc,char**argv,char**envp)
  {
  char bufr[128],*try=" ...try\nClinch inputfile";
  char*output="Clinch.csv";
  int i,hour;
  double dt,qdam,qinflo,t,t0,tin,tinflo;
  FILE*f5,*f6;
  printf("Clinch River Flow Routing Model\n");
  printf("solving kinematic wave equation with temperature\n");
  if(argc<2)
    Abort(__LINE__,"no input file%s",try);
  if(argc>2)
    Abort(__LINE__,"too many arguments%s",try);
  printf("input file: %s\n",argv[1]);
  if((f5=fopen(argv[1],"rt"))==NULL)
    Abort(__LINE__,"can't open input file");
  printf("output file: %s\n",output);
  if((f6=fopen(output,"wt"))==NULL)
    Abort(__LINE__,"can't create output file");
  hour=0;
  t=0.;
  dt=1./ndt;
  while(fgets(bufr,sizeof(bufr),f5))
    {
    if(sscanf(bufr,"%lf %lf %lf",&qdam,&tin,&t0)!=3)
      {
      printf("scan error\n");
      break;
      }
    if((++hour)%24==1)
      printf("routing day %i\n",(hour+23)/24);
    for(i=0;i<ndt;i++)
      {
      int err;
      if((err=Routing(qdam*100.,tin,t0,ndt,&qinflo,&tinflo))!=0)
        {
        printf("error code %i\n",err);
        goto done;
        }
      t+=dt;
      fprintf(f6,"%lG,%lG,%lG,%lG,%lG,%lG\n",t,qdam,qinflo/100.,tin,tinflo,t0);
      }
    }
done:
  fclose(f6);
  fclose(f5);
  return(0);
  }
