/* Wheeler Reservoir Flow Routing
   using the MacCormack method
   written in FORTRAN in 1982
   dudley.benton@gmail.com
*/
#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

void Abort(int line,char*format,...)
  {
  va_list arg_marker;
  va_start(arg_marker,format);
  fprintf(stderr,"\nerror %i: ",line);
  vfprintf(stderr,format,arg_marker);
  fprintf(stderr,"\n");
  _fcloseall();
  exit(1);
  }

double fmin(double x,double y)
  {
  if(x<y)
    return(x);
  return(y);
  }

double fmax(double x,double y)
  {
  if(x>y)
    return(x);
  return(y);
  }

double sq(double x)
  {
  return(x*x);
  }

int LeapYear(int year)
  {
  if(year>=1582)
    {
    if((year%4)==0)
      if((year%100)!=0||(year%400)==0)
        return(1);
    return(0);
    }
  if(year>0)
    {
    if((year%4)==0)
      return(1);
    return(0);
    }
  if((year+4713)%4==0)
    return(1);
  return(0);
  }

int JulianDay(int year,int month,int day)
  {
  int a,m,y;
  if(year<-4713||year==0)
    return(-999);
  if(month<1||month>12||day<1||day>31)
    return(-999);
  if(month==2&&day==29)
    if(!LeapYear(year))
      return(-999);
  if(month==2&&day>29)
    return(-999);
  if(month==4||month==6||month==9||month==11)
    if(day>30)
      return(-999);
  if(year==1582)
    if(month==10)
      if(5<=day&&day<=14)
        return(-999);
  if(year<0)
    year+=1;
  a=(14-month)/12;
  y=year+4800-a;
  m=month+12*a-3;
  if((year==1582&&month==10&&day<=4)||(year<1582))
    return(day+(153*m+2)/5+365*y+y/4-32083);
  return(day+(153*m+2)/5+365*y+y/4-y/100+y/400-32045);
  }

int MicrosoftDay(int year,int month,int day)
  {
  return(JulianDay(year,month,day)-2415019);
  }

/* river geometry */

#define ns  8 /* number of sections */
#define ne  6 /* number of elevations */
#define ndt 6 /* time steps per hour */
#define ndi 2 /* number of initialization days */

double area[ns+1][ne]={
    1093.,  8977., 19193., 30081., 54006., 97896.,
    4123., 14206., 25681., 37963., 66721., 11795.,
    3961., 14884., 27025., 46055., 97208.,171227.,
    4512., 17615., 32119., 67591.,141693.,236414.,
    5687., 21182., 42266.,115346.,206595.,308148.,
   11165., 29614.,102096.,217524.,350564.,495495.,
   15839., 53464.,132986.,217914.,311498.,420312.,
  101143.,167535.,236733.,307466.,488231.,582693.,
  218708.,298970.,379714.,460941.,476589.,624850.};
double r23[ns+1][ne]={
  0.82, 4.34, 6.13, 8.32, 8.54, 7.78,
  3.00, 5.68, 7.96, 9.86, 9.46, 8.52,
  2.70, 5.51, 7.55, 8.01, 7.24, 6.77,
  2.27, 5.49, 9.50, 7.19, 5.96, 5.70,
  2.45, 5.61, 7.08, 6.32, 6.40, 6.82,
  3.70, 6.22, 5.96, 7.25, 8.90, 9.92,
  3.70, 5.47, 6.94, 9.01,10.69,11.93,
  6.44, 7.02,10.68,12.53,14.19,15.63,
  9.13,11.14,11.76,13.57,15.97,17.86};
double width[ns+1][ne]={
  1.,1122., 1479., 3667., 6611., 8811.,
  1.,1047., 1931., 4700.,13583.,17127.,
  1.,1055., 1432., 3272., 8117.,11894.,
  1.,1523., 5225.,14206.,25378.,30394.,
  1.,2238., 9131.,21222.,28696.,33056.,
  1.,4078.,11750.,17122.,25922.,35429.,
  1.,7006., 9896.,12567.,17283.,23273.,
  1.,9577.,12786.,15582.,17418.,17790.,
  1.,7832., 8025., 8477., 9106., 9144.};
double manning[ns+1]={0.0265,0.024,0.024,0.025,0.026,0.028,0.03,0.03,0.03};
double dx=48840.;

/* flow routing arrays */

int init=1;
int nday;
int new;
double a[ns+1];
double b[ns+1];
double h[2][ns];
double phpt[2][ns];
double pqpt[2][ns+1];
double qin[ns];
double qqa[ns+1];
double q[2][ns+1];
double s[ns+1];
double sf[ns];

void Initialize(double twup,double hwdn,double qup,double qdn)
  {
  int i,j;
  for(i=0;i<ns;i++)
    for(j=0;j<2;j++)
      h[j][i]=hwdn+(twup-hwdn)*sq(sq((ns-i-1)/7.));
  for(i=0;i<=ns;i++)
    for(j=0;j<2;j++)
      q[j][i]=qup+(qdn-qup)*i/8.;
  }

void Geometry(double twup,double hwdn,double qup,double qdn,int new)
  {
  int i,i1,i2,j,j1,k,l;
  double hbar,r23t,x,xj;
  for(l=0;l<2;l++)
    {
    for(k=i=0;i<ns+1;i++)
      {
      i1=max(1,i);
      i2=min(ns,i+1);
      hbar=(h[new][i1-1]+h[new][i2-1])/2.;
      x=(hbar-525.)/10.;
      if(x<1.||x>5.)
        k++;
      j=max(1,min(ne-1,(int)x))-1;
      j1=j+1;
      xj=x-j1;
      a[i]=area[i][j]+(area[i][j1]-area[i][j])* xj;
      b[i]=width[i][j]+(width[i][j1]-width[i][j])*xj;
      r23t=r23[i][j]+(r23[i][j1]-r23[i][j])*xj;
      s[i]=sq(manning[i]/(1.486*a[i]*r23t));
      }
    if(!k)
      return;
    Initialize(twup,hwdn,qup,qdn);
    }
  Abort(__LINE__,"reservior elevations have become unstable");
  }

#define gravity 32.174

void Routing() /* MacCormack method */
  {
  int i;
  double bbar,phpx,pqpx,pqqapx;
  for(i=0;i<=ns;i++)
    qqa[i]=sq(q[new][i])/a[i];
  for(i=0;i<ns;i++)
    {
    pqpx=(q[new][i+1]-q[new][i])/dx;
    bbar=(b[i]+b[i+1])/2.;
    phpt[new][i]=(qin[i]/dx-pqpx)/bbar;
    }
  for(i=1;i<ns;i++)
    {
    phpx=(h[new][i]-h[new][i-1])/dx;
    pqqapx=(qqa[i+1]-qqa[i-1])/(dx+dx);
    pqpt[new][i]=-gravity*a[i]*(phpx+sf[i])-pqqapx;
    }
  }

char bufr[1024];
FILE*f5;
FILE*f6;

void River(int*month,int*iday,int*iyear,int*ihour,double*qup,double*qdn,double*twup,double*hwdn)
  {
  int i,idays,jhour,istep,iter,ndays,nhour,nstep;
  double dq,dqmax,dt;
  dt=3600./ndt;
  if(init)
    Initialize(*twup,*hwdn,qup[0],qdn[0]);
  ndays=ndi+1;
  nhour=24;
  nstep=ndt;
  if(init<=0)
    {
    ndays=1;
    if(init<0)
      {
      nhour=1;
      nstep=-init;
      }
    }
  for(idays=0;idays<ndays;idays++)
    {
    for(jhour=0;jhour<nhour;jhour++)
      {
      for(istep=0;istep<nstep;istep++)
        {
        new=1-new;
        q[new][0]=qup[jhour];
        q[new][ns]=qdn[jhour];
        for(i=0;i<ns;i++)
          {
          h[new][i]=h[1-new][i]+dt*phpt[1-new][i]/2.;
          h[new][i]=fmax(535.,fmin(585.,h[new][i]));
          }
        for(i=1;i<ns;i++)
          q[new][i]=q[1-new][i]+dt*pqpt[1-new][i]/2.;
        for(iter=0;iter<20;iter++)
          {
          Geometry(*twup,*hwdn,qup[jhour],qdn[jhour],new);
          for(i=1;i<ns;i++)
            sf[i]=s[i]*q[new][i]*fabs(q[new][i]);
          Routing(new);
          for(i=0;i<ns;i++)
            {
            h[new][i]=h[1-new][i]+dt*(phpt[1-new][i]+phpt[new][i])/2.;
            h[new][i]=fmax(535.,fmin(585.,h[new][i]));
            }
          dqmax=0.;
          for(i=1;i<ns;i++)
            {
            dq=q[new][i];
            q[new][i]=q[1-new][i]+dt*(pqpt[1-new][i]+pqpt[new][i])/2.;
            dq=fabs(q[new][i]-dq);
            dqmax=fmax(dqmax,dq);
            }
          if(dqmax<100.)
            break;
          }
        if(idays<ndays-1)
          continue;
        fprintf(f6,"%lf",MicrosoftDay(iyear[0],month[0],iday[0])+(jhour+istep/(double)nstep)/(double)nhour);
        for(i=0;i<ns;i++)
          fprintf(f6,",%lG",h[new][i]);
        for(i=0;i<=ns;i++)
          fprintf(f6,",%lG",q[new][i]);
        fprintf(f6,"\n");
        }
      }
    }
  }

enum{err_eof=1,err_scan,err_mode};

int ReadInput(int*month,int*iday,int*iyear,int*ihour,double*qup,double*qdn)
  {
  int i;
  static double hwdn,twup;
  if(init)
    {
    fgets(bufr,sizeof(bufr),f5);
    if(sscanf(bufr,"%lf,%lf",&twup,&hwdn)!=2)
      Abort(__LINE__,"scan error on initial conditions");
    }
  for(i=0;i<24;i++)
    {
    if(!fgets(bufr,sizeof(bufr),f5))
      return(err_eof);
    if(sscanf(bufr,"%d/%d/%d %d:%*d,%lf,%lf",
        &month[i],&iday[i],&iyear[i],&ihour[i],&qup[i],&qdn[i])!=6)
      return(err_scan);
    qup[i]*=100.;
    qdn[i]*=100.;
    }
  nday++;
  printf("day %d  input complete\n  begin flow routing\n",nday);
  River(month,iday,iyear,ihour,qup,qdn,&twup,&hwdn);
  printf("  flow routing complete\n");
  init=0;
  return(0);
  }

int main(int argc,char**argv,char**envp)
  {
  char*try="...\ntry Wheeler inputfile\n";
  char*oname="Wheeler.csv";
  int iday[24],ier=0,ihour[24],iyear[24],month[24];
  double qup[24],qdn[24];
  printf("Wheeler Reservoir Flow Routing\n");
  if(argc<2)
    {
    printf("no input file specified%s",try);
    return(1);
    }
  if(argc>2)
    {
    printf("too many parameters%s",try);
    return(1);
    }
  printf("input file: %s\n",argv[1]);
  if((f5=fopen(argv[1],"rt"))==NULL)
    {
    printf("can't open input file\n");
    return(1);
    }
  printf("output file: %s\n",oname);
  if((f6=fopen(oname,"wt"))==NULL)
    {
    printf("can't create output file\n");
    return(1);
    }
  while(ReadInput(month,iday,iyear,ihour,qup,qdn)==0)
    continue;
  fclose(f5);
  fclose(f6);
  return(0);
  }
