/*******************************************************************************

                    These moist air property functions were
                               developed by
                             Dudley J. Benton
                            Knoxville, Tennessee
                           dudley.benton@gmail.com

  These properties are based on the 1993 ASHRAE Handbook of Fundamentals.
  The tables in the Handbook are derived from the work of Hyland & Wexlar.

  Note 1: The formulas in the Handbook do not produce the values in these
          tables; however, these functions do. The tables are correct. The
          formulas in the Handbook are not.

  Note 2: Relative humidity is a number between 0 and 1, NOT between 0 and 100.

  Note 3: ASHRAE defines the specific volume of moist air as cubic feet per
          pound of DRY air, NOT per pound of moist (or total) air. Use function
          fVol(P,Tdb,W) for this. If you want cubic feet per pound of moist
          (or total) air, divide this by 1+W.

  Note 4: The function fRho(P,Tdb,W) returns pounds of MOIST (or total) air
          per cubic foot. If you want pounds of DRY air per cubic foot, use
          1/fVol(P,Tdb,W).

  Note 5: ASHRAE defines the enthalpy of moist air in BTU per pound of DRY
          air, NOT per pound of moist (or total) air. If you want BTU per
          pound of moist air multiply this by 1+W.

  Note 6: In spite of the fact that there are many incorrect references,
          H=Ha+W*Hg, NOT H=Ha+W*Hfg! ASHRAE has it correctly. If you're still
          convinced that Hfg is the correct term, rather than Hg, consider
          the case where you're mixing steam at the critical point with
          air. Lots of energy, but Hfg=0 at the critical point. Hfg is
          simply wrong. Don't use it.

  02/12/2018: replace fT(Ps,Ws) calculation with curve-fit
              replace fWs(T) curve-fit with calculation

*******************************************************************************/

#define _CRT_SECURE_NO_DEPRECATE
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

double fT(double degF) /* temperaure dependence of the enhancement factor */
  {
  return(((((-1.22884575824E-13*degF+2.98501566007E-11)*degF
    -1.68742180749E-9)*degF+1.64960460192E-7)*degF
    -1.37450341076E-5)*degF+1.00432811563436);
  }

double fP(double degF,double psia) /* pressure dependence of enhancement factor */
  {
  return(0.9999278133+(0.00002200673657
    +(6.729181914E-8-7.590092091E-10*degF)*degF)*degF
    +(0.0002997989681+(-0.000001718254802+8.054776534E-9*degF)*degF
    -3.40732948E-9*degF*psia)*psia);
  }

double fPT(double psia,double degF) /* composite (P and T) enhancement factor */
  {
  return(fT(degF)*fP(degF,psia)/fP(degF,14.696));
  }

double fPs(double degF) /* saturation pressure (-160F to +200F) */
  {
  double degR;
  degR=degF+459.67;
  if(degF<=32.)
    return(exp(((((-9.0344688E-14*degR+3.5575832E-10)*degR
      +0.00000019202377)*degR-0.0053765794)*degR-4.8932428)
      -10214.165/degR+4.1635019*log(degR)));
  return(exp((((-2.4780681E-9*degR+0.00001289036)*degR-0.027022355)*degR
    -11.29465)-10440.397/degR+6.5459673*log(degR)));
  }

double fTs(double psia) /* saturation temperature from pressure */
  {
  int iter;
  double degF,T1,T2;
  T1=-160.;
  T2=200.;
  for(iter=0;iter<32;iter++)
    {
    degF=(T1+T2)/2.;
    if(fPs(degF)<psia)
      T1=degF;
    else
      T2=degF;
    }
  return(degF);
  }

double fWs(double psia,double degF) /* humidity ratio at saturation */
  {
  double f,Ps;
  f=fPT(psia,degF);
  Ps=fPs(degF);
  return(0.62198*f*Ps/(psia-f*Ps));
  }

double fRHdbw(double P,double Tdb,double W) /* RH from Tdb and W */
  {
  double f,Ps;
  f=fPT(P,Tdb);
  Ps=fPs(Tdb);
  return(W*P/(f*Ps*(W+0.62198)));
  }

double fWdbrh(double baro,double Tdb,double RH) /* W from Tdb and RH */
  {
  double f,Ps;
  if(RH<=0.)
    return(0.);
  f=fPT(baro,Tdb);
  Ps=fPs(Tdb);
  if(RH>=1.)
    return(0.62198/(baro/(f*Ps)-1.));
  return(0.62198/(baro/(RH*(f*Ps))-1.));
  }

double fRHdbp(double P,double Tdb,double Tdp) /* RH from Tdb and Tdp */
  {
  if(Tdb<=Tdp)
    return(1.);
  return((fPT(P,Tdp)*fPs(Tdp))/(fPT(P,Tdb)*fPs(Tdb)));
  }

double fWdbp(double P,double Tdb,double Tdp) /* W from Tdb and Tdp */
  {
  double f,Ps;
  f=fPT(P,Tdp);
  Ps=fPs(Tdp);
  return(0.62198/(P/(f*Ps)-1.));
  }

double fTdw(double P,double W) /* Tdp fron P and W */
  {
  int iter;
  double T1,T2,Tdp;
  T1=-160.;
  T2=200.;
  for(iter=0;iter<32;iter++)
    {
    Tdp=(T1+T2)/2.;
    if(fWdbrh(P,Tdp,1.)<W)
      T1=Tdp;
    else
      T2=Tdp;
    }
  return(Tdp);
  }

double fTdprh(double P,double Tdb,double RH) /* Tdp from Tdb and RH */
  {
  return(fTdw(P,fWdbrh(P,Tdb,RH)));
  }

double fHa(double Tdb) /* enthalpy of dry air */
  {
  return(((9.70200247051023E-9*Tdb+6.6508218793393E-8)*Tdb
     +0.240202334543101)*Tdb+1.61183146829558E-5);
  }

double fHg(double Tdb) /* enthalpy of water vapor */
  {
  return(((-4.97737026483591E-7*Tdb+2.56528645046217E-5)*Tdb
    +0.43936852944779)*Tdb+1060.69129737909);
  }

double fHdbw(double Tdb,double W) /* H from Tdb and W */
  {
  return(fHa(Tdb)+W*fHg(Tdb));
  }

double fHdbrh(double P,double Tdb,double RH) /* H from Tdb and RH */
  {
  return(fHdbw(Tdb,fWdbrh(P,Tdb,RH)));
  }

double fWdwb(double P,double Tdb,double Twb) /* W from Tdb and Twb */
  {
  double Ws;
  Ws=fWdbrh(P,Twb,1.);
  if(Tdb<=Twb)
    return(Ws);
  return((fHa(Twb)-fHa(Tdb)+Ws*fHg(Twb))/fHg(Tdb));
  }

double fHdwb(double P,double Tdb,double Twb) /* H from Tdb and Twb */
  {
  return(fHdbw(Tdb,fWdwb(P,Tdb,Twb)));
  }

double fTdbhw(double P,double H,double W) /* Tdb from H and w */
  {
  int iter;
  double T1,T2,Tdb;
  T1=-160.;
  T2=200.;
  for(iter=0;iter<32;iter++)
    {
    Tdb=(T1+T2)/2.;
    if(fHdbw(Tdb,W)<H)
      T1=Tdb;
    else
      T2=Tdb;
    }
  return(Tdb);
  }

double fTdwbw(double P,double Twb,double W) /* Tdb from Twb and W */
  {
  int iter;
  double T1,T2,Tdb;
  T1=Twb;
  T2=200.;
  for(iter=0;iter<32;iter++)
    {
    Tdb=(T1+T2)/2.;
    if(fWdwb(P,Tdb,Twb)>W)
      T1=Tdb;
    else
      T2=Tdb;
    }
  return(Tdb);
  }

double fRHdwb(double P,double Tdb,double Twb) /* RH from Tdb and Twb */
  {
  if(Tdb<=Twb)
    return(1.);
  return(fRHdbw(P,Tdb,fWdwb(P,Tdb,Twb)));
  }

double fTdwrh(double P,double Twb,double RH) /* Tdb from Twb and RH */
  {
  int iter;
  double T1,T2,Tdb;
  T1=Twb;
  T2=200.;
  for(iter=0;iter<32;iter++)
    {
    Tdb=(T1+T2)/2.;
    if(fRHdwb(P,Tdb,Twb)>RH)
      T1=Tdb;
    else
      T2=Tdb;
    }
  return(Tdb);
  }

double fTwdrh(double P,double Tdb,double RH) /* Twb from Tdb and RH */
  {
  int iter;
  double T1,T2,Twb;
  T1=-160.;
  T2=200.;
  for(iter=0;iter<32;iter++)
    {
    Twb=(T1+T2)/2.;
    if(fRHdwb(P,Tdb,Twb)<RH)
      T1=Twb;
    else
      T2=Twb;
    }
  return(Twb);
  }

double fWbdbw(double P,double Tdb,double W) /* Twb from Tdb and W */
  {
  int iter;
  double T1,T2,Twb;
  T1=-160.;
  T2=200.;
  for(iter=0;iter<32;iter++)
    {
    Twb=(T1+T2)/2.;
    if(fWdwb(P,Tdb,Twb)<W)
      T1=Twb;
    else
      T2=Twb;
    }
  return(Twb);
  }

double fTdpdw(double P,double Tdb,double Twb) /* Tdp from Tdb and Twb */
  {
  return(fTdw(P,fWdwb(P,Tdb,Twb)));
  }

double fTwdbp(double P,double Tdb,double Tdp) /* Twb from Tdb and Tdp */
  {
  int iter;
  double T1,T2,Twb;
  T1=Tdp;
  T2=Tdb;
  for(iter=0;iter<32;iter++)
    {
    Twb=(T1+T2)/2.;
    if(fTdpdw(P,Tdb,Twb)<Tdp)
      T1=Twb;
    else
      T2=Twb;
    }
  return(Twb);
  }

double fTdwbp(double P,double Twb,double Tdp) /* Tdb from Twb and Tdp */
  {
  int iter;
  double T1,T2,Tdb;
  T1=Tdp;
  T2=200.;
  for(iter=0;iter<32;iter++)
    {
    Tdb=(T1+T2)/2.;
    if(fTdpdw(P,Tdb,Twb)>Tdp)
      T1=Tdb;
    else
      T2=Tdb;
    }
  return(Tdb);
  }

double tcv(double T) /* temperature correction for specific volume */
  {
  return(((((((3.800681764821E-16*T-2.547540532089E-13)*T
    +6.65477336779E-11)*T-8.77345005328E-9)*T+6.702472798597E-7)*T
    -3.728540388556E-5)*T+0.001603091098844)*T+1.570844459163);
  }

double fVol(double P,double Tdb,double W) /* cu.ft./lbm DRY air */
  {
  return(53.352*(Tdb+459.67)*(1.+tcv(Tdb)*W)/P/144.);
  }

double fRho(double P,double Tdb,double W) /* lbm MOIST air/cu.ft. */
  {
  return((1.+W)/fVol(P,Tdb,W));
  }
