/* Solar Collector Field Simulation
   developed by dudley.benton@gmail.com
*/
#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

double RE=6371.01;    /* radius of Earth */
double AU=149597890.; /* astronomical unit */

typedef struct{double z,a;}SP;

SP SolarPosition(int yr,int mo,int da,int hr,int mn,int sec,double lon,double lat)
  {
  int i,j;
  double a,b,c,d,e,g,h,l,o,r,s,t,w;
  static SP sp;
  h=hr+(mn+sec/60.)/60.;
  i=(mo-14)/12;
  j=(1461*(yr+4800+i))/4+(367*(mo-2-12*i))/12-(3*((yr+4900+i)/100))/4+da-32075;
  s=j-0.5+h/24.;
  d=s-2451545.;
  w=2.1429-0.0010394594*d;
  l=4.895063+0.017202791698*d;
  b=6.24006+0.0172019699*d;
  e=l+0.03341607*sin(b)+0.00034894*sin(2.*b)-0.0001134-0.0000203*sin(w);
  o=0.4090928-0.000000006214*d+0.0000396*cos(w);
  r=atan2(cos(o)*sin(e),cos(e));
  if(r<0.)
    r=r+2.*M_PI;
  c=asin(sin(o)*sin(e));
  g=6.6974243242+0.0657098283*d+h;
  t=(g*15.+lon)*(M_PI/180.);
  a=t-r;
  sp.z=(acos(cos(lat*(M_PI/180.))*cos(a)*cos(c)+sin(c)*sin(lat*(M_PI/180.))));
  sp.a=atan2(-sin(a),tan(c)*cos(lat*(M_PI/180.))-sin(lat*(M_PI/180.))*cos(a));
  if(sp.a<0.)
    sp.a=sp.a+2.*M_PI;
  sp.a/=M_PI/180.;
  sp.z=(sp.z+(RE/AU)*sin(sp.z))/(M_PI/180.);
  return(sp);
  }

void TestSolarPosition()
  {
  static int da,hr,mn,mo,sec,yr=2019;
  double lat=40.7128,lon=-74.006;
  SP sp;
  for(da=1;da<=3;da++)
    {
    for(hr=0;hr<24;hr++)
      {
      sp=SolarPosition(yr,mo,da,hr,mn,sec,lon,lat);
      printf("%02i/%02i/%04i %2i:%02i:%02i %3.0lf %3.0lf\n",mo,da,yr,hr,mn,sec,sp.z,sp.a);
      }
    }
  }

int LeapYear(int year)
  {
  if(year>=1582)
    {
    if((year%4)==0)
      if((year%100)!=0||(year%400)==0)
        return(1);
    return(0);
    }
  if(year>0)
    {
    if((year%4)==0)
      return(1);
    return(0);
    }
  if((year+4713)%4==0)
    return(1);
  return(0);
  }

int JulianDay(int year,int month,int day)
  {
  int a,m,y;
  if(year<-4713||year==0)
    return(-999);
  if(month<1||month>12||day<1||day>31)
    return(-999);
  if(month==2&&day==29)
    if(!LeapYear(year))
      return(-999);
  if(month==2&&day>29)
    return(-999);
  if(month==4||month==6||month==9||month==11)
    if(day>30)
      return(-999);
  if(year==1582)
    if(month==10)
      if(5<=day&&day<=14)
        return(-999);
  if(year<0)
    year+=1;
  a=(14-month)/12;
  y=year+4800-a;
  m=month+12*a-3;
  if((year==1582&&month==10&&day<=4)||(year<1582))
    return(day+(153*m+2)/5+365*y+y/4-32083);
  return(day+(153*m+2)/5+365*y+y/4-y/100+y/400-32045);
  }

void YearMonthDay(int jday,int*year,int*month,int*day)
  {
  year[0]=(int)(jday/365.25)-4713;
  while(JulianDay(year[0],1,1)>jday)
    year[0]-=1;
  while(JulianDay(year[0],12,31)<jday)
    year[0]+=1;
  month[0]=12;
  while(JulianDay(year[0],month[0],1)>jday)
    month[0]-=1;
  day[0]=jday-JulianDay(year[0],month[0],1)+1;
  }

int MicrosoftDay(int year,int month,int day)
  {
  return(JulianDay(year,month,day)-2415019);
  }

double DNIearly(int aday)
  {
  return(384.4+63.06*cos(2.*M_PI*(aday-363)/365.25));
  }

double DNIlate(int aday)
  {
  return(1111.2+68.14*cos(2.*M_PI*(aday-165)/365.25));
  }

double DNIpeak(int aday)
  {
  return(980.*(1.+0.033412*cos(2.*M_PI*(aday-64)/365.24)));
  }

double DNIclear(int aday,int hour,int minute)
  {
  double k=0.163,early,late,noon,peak,theta,time;
  early=DNIearly(aday);
  late=DNIlate(aday);
  noon=(early+late)/2.;
  peak=DNIpeak(aday);
  time=hour*60+minute;
  theta=M_PI*(time-noon)/(late-early);
  if(fabs(theta)<0.999*M_PI/2.)
    return(peak*exp(k)/exp(k/cos(theta)));
  return(0.);
  }

double Cloudiness(double probability)
  {
  double c;
  c=((((0.76288721101*probability-1.28320573162)*probability
    +0.624789144234)*probability-0.0558391005561)*probability
    -0.000509558186569)/((1.25994765396*probability-2.21186907023)
    *probability+1.);
  if(c<0.)
    return(0.);
  if(c>1.)
    return(1.);
  return(c);
  }

void Simulation(char*fname)
  {
  int day,hour,julian,minute,mo,month=0,y=0,year,yr;
  double area=75.,cloudiness=0.,DNI,MW,MWhr=0.,t=1./60./3.;
  FILE*fp;
  printf("Solar Collector Field Simulation\n");
  printf("output file: %s\n",fname);
  if((fp=fopen(fname,"wt"))==NULL)
    return;
  for(year=2019;year<2020;year++)
    {
    for(julian=JulianDay(year,1,1);julian<=JulianDay(year,12,31);julian++)
      {
      mo=month;
      yr=y;
      YearMonthDay(julian,&y,&month,&day);
      if(day==1&&mo>0)
        {
        printf("%i/%i,%.0lf\n",mo,yr,MWhr);
        MWhr=0.;
        }
      for(hour=0;hour<24;hour++)
        {
        for(minute=0;minute<60;minute+=15)
          {
          DNI=DNIclear(julian-JulianDay(year,1,1),hour,minute);
          cloudiness=(1.-t)*cloudiness+t*Cloudiness(rand()/32767.);
          MW=area*DNI*cloudiness;
          MWhr+=MW/60./1000.;
          fprintf(fp,"%i/%i/%i %i:%02i,%.0lf\n",month,day,year,hour,minute,MW);
          }
        }
      }
    }
  printf("%i/%i,%.0lf\n",mo,yr,MWhr);
  fclose(fp);
  }

int main(int argc,char**argv,char**envp)
  {
//TestSolarPosition();
  Simulation("solar.csv");
  return(0);
  }
